/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.shardingsphere.sql.parser.integrate.engine;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.shardingsphere.sql.parser.SQLParserEngineFactory;
import org.apache.shardingsphere.sql.parser.integrate.asserts.SQLCaseAssertContext;
import org.apache.shardingsphere.sql.parser.integrate.asserts.statement.SQLStatementAssert;
import org.apache.shardingsphere.sql.parser.integrate.jaxb.SQLParserTestCasesRegistry;
import org.apache.shardingsphere.sql.parser.integrate.jaxb.SQLParserTestCasesRegistryFactory;
import org.apache.shardingsphere.sql.parser.integrate.jaxb.domain.statement.SQLParserTestCase;
import org.apache.shardingsphere.sql.parser.sql.statement.SQLStatement;
import org.apache.shardingsphere.test.sql.SQLCaseType;
import org.apache.shardingsphere.test.sql.loader.SQLCasesLoader;
import org.apache.shardingsphere.test.sql.loader.SQLCasesRegistry;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;

import static org.junit.Assert.fail;

@RunWith(Parameterized.class)
@RequiredArgsConstructor
@Slf4j
public final class SQLParserParameterizedTest {
    
    private static final SQLCasesLoader SQL_CASES_LOADER = SQLCasesRegistry.getInstance().getSqlCasesLoader();
    
    private static final SQLParserTestCasesRegistry SQL_PARSER_TEST_CASES_REGISTRY = SQLParserTestCasesRegistryFactory.getInstance().getRegistry();
    
    private final String sqlCaseId;
    
    private final String databaseType;
    
    private final SQLCaseType sqlCaseType;
    
    @Parameters(name = "{0} ({2}) -> {1}")
    public static Collection<Object[]> getTestParameters() {
        // TODO resume me after all test cases passed 
//        checkTestCases();
        return getSQLTestParameters();
    }
    
    private static void checkTestCases() {
        Collection<String> allSQLCaseIDs = new HashSet<>(SQL_CASES_LOADER.getAllSQLCaseIDs());
        if (allSQLCaseIDs.size() != SQL_PARSER_TEST_CASES_REGISTRY.getAllSQLCaseIDs().size()) {
            allSQLCaseIDs.removeAll(SQL_PARSER_TEST_CASES_REGISTRY.getAllSQLCaseIDs());
            fail(String.format("The count of SQL cases and SQL parser cases are mismatched, missing cases are: %s", allSQLCaseIDs));
        }
    }
    
    private static Collection<Object[]> getSQLTestParameters() {
        Collection<Object[]> result = new LinkedList<>();
        // TODO resume me after all test cases passed 
//        for (Object[] each : SQL_CASES_LOADER.getSQLTestParameters()) {
        for (Object[] each : getSQLTestParameters(SQL_CASES_LOADER.getSQLTestParameters())) {
            if (!isPlaceholderWithoutParameter(each)) {
                result.add(each);
            }
        }
        return result;
    }
    
    // TODO remove me after all test cases passed
    private static Collection<Object[]> getSQLTestParameters(final Collection<Object[]> sqlTestParameters) {
        Collection<Object[]> result = new LinkedList<>();
        for (Object[] each : sqlTestParameters) {
            if (!isPassedSqlCase(each[0].toString())) {
                result.add(each);
            }
        }
        return result;
    }
    
    private static boolean isPassedSqlCase(final String sqlCaseId) {
        Collection<String> sqlCases = new LinkedList<>();
        sqlCases.add("show_index_with_indexes_with_table_and_database");
        sqlCases.add("show_index_with_database_back_quotes");
        sqlCases.add("show_index_with_table_back_quotes");
        // TODO Stop index is wrong
        sqlCases.add("select_with_expression");
        // TODO Sub query is necessary
        sqlCases.add("select_pagination_with_row_number");
        sqlCases.add("select_pagination_with_row_number_and_diff_group_by_and_order_by");
        sqlCases.add("select_pagination_with_row_number_and_diff_group_by_and_order_by");
        sqlCases.add("select_pagination_with_row_number_and_group_by_and_order_by");
        sqlCases.add("select_pagination_with_row_number_for_greater_than");
        sqlCases.add("select_pagination_with_row_number_for_greater_than_and_equal");
        sqlCases.add("select_pagination_with_offset_fetch");
        sqlCases.add("select_pagination_with_top");
        sqlCases.add("select_pagination_with_top_for_greater_than");
        sqlCases.add("select_pagination_with_top_for_greater_than_and_equal");
        sqlCases.add("select_pagination_with_top_and_group_by_and_order_by");
        sqlCases.add("select_pagination_with_top_and_group_by_and_order_by_and_parentheses");
        sqlCases.add("select_pagination_with_top_and_diff_group_by_and_order_by");
        sqlCases.add("select_pagination_with_top_and_diff_group_by_and_order_by_and_parentheses");
        // TODO Alter statement needs new segment
        sqlCases.add("alter_table_add_foreign_key");
        sqlCases.add("alter_table_add_primary_foreign_key");
        sqlCases.add("alter_table_add_constraints_sqlserver");
        // TODO cannot parse create index behind pk in create table statement, and new segment is necessary
        sqlCases.add("create_table_with_create_index");
        sqlCases.add("create_table_with_exist_index");
        // TODO cannot support insert all
        sqlCases.add("insert_all_with_all_placeholders");
        return sqlCases.contains(sqlCaseId);
    }
    
    private static boolean isPlaceholderWithoutParameter(final Object[] sqlTestParameter) {
        return SQLCaseType.Placeholder == sqlTestParameter[2] && SQL_PARSER_TEST_CASES_REGISTRY.get(sqlTestParameter[0].toString()).getParameters().isEmpty();
    }
    
    @Test
    public void assertSupportedSQL() {
        SQLParserTestCase expected = SQL_PARSER_TEST_CASES_REGISTRY.get(sqlCaseId);
        String databaseType = "H2".equals(this.databaseType) ? "MySQL" : this.databaseType;
        String sql = SQL_CASES_LOADER.getSQL(sqlCaseId, sqlCaseType, SQL_PARSER_TEST_CASES_REGISTRY.get(sqlCaseId).getParameters());
        SQLStatement actual = SQLParserEngineFactory.getSQLParserEngine(databaseType).parse(sql, false);
        SQLStatementAssert.assertIs(new SQLCaseAssertContext(sqlCaseId, sqlCaseType), actual, expected);
    }
}
